<?php

use \GuzzleHttp\Exception\RequestException;
use \GuzzleHttp\Psr7;
use Symfony\Component\Yaml\Yaml;

/**
 * @file
 * Entity class definition for PiftCiJob class.
 */
class PiftCiJob extends Entity {

  /**
   * @var int Unique ID for this CI data entity.
   */
  public $job_id;

  /**
   * @var string Environment type for the test.
   */
  public $environment;

  /**
   * @var string Type of test, PIFT_CI_TYPE_*.
   */
  public $target_type;

  /**
   * @var string Related CI detail URL for results.
   */
  public $ci_url;

  /**
   * @var string Status of the test, PIFT_CI_STATUS_*.
   */
  public $status = PIFT_CI_STATUS_QUEUE;

  /**
   * @var string Result of the test, PIFT_CI_RESULT_*.
   */
  public $result = NULL;

  /**
   * @var string Summary message of test result.
   */
  public $message = 'Queueing';

  /**
   * @var int Issue node ID.
   */
  public $issue_nid;

  /**
   * @var int Related CI detail record ID user for retrieving results. Only used when type is PIFT_CI_TYPE_FILE.
   */
  public $file_id;

  /**
   * @var string Issue branch when type is PIFT_CI_TYPE_MERGE_REQUEST.
   */
  public $issue_branch = NULL;

  /**
   * @var string The branch that a merge request will be merged into.
   */
  public $merge_target = NULL;

  /**
   * @var int Release node ID.
   */
  public $release_nid;

  /**
   * @var string Core branch to test with. Only used when test is for a
   * non-core project.
   */
  public $core_branch = '';

  /**
   * @var string SHA1 of the commit that triggered this test. Used when type is
   *   PIFT_CI_TYPE_BRANCH or PIFT_CI_TYPE_MERGE_REQUEST.
   */
  public $commit_id;

  /**
   * @var int Number of times this build (or test) has been requested.
   */
  public $build_iteration = 1;

  /**
   * @var int indicating a jobs order in the testing queue.
   */
  public $job_priority;

  /**
   * @var int Timestamp when this record was updated.
   */
  public $updated = NULL;

  /**
   * @var int Timestamp when this record was created.
   */
  public $created = NULL;

  /**
   * @var int User ID of the user who created this job.
   */
  public $uid = NULL;

  /**
   * @var string Reason the job was created.
   */
  public $reason = '';

  /**
   * @var int User ID of the user who requested bypassing a failing branch.
   */
  public $bypass_uid = NULL;

  /**
   * @var bool Flag for whether a test was custom or standard.
   */
  public $custom = 0;

  /**
   * @var string The category indicator for this test: legacy/current
   */
  private $release_category = NULL;

  /**
   * @var string Extended build details, usually failure explanation.
   */
  public $build_details = '';

  /**
   * @var int Count of checkstyle messages. The actual messages may be
   *   agressively cleaned.
   */
  public $checkstyle = 0;

  /**
   * @var int If job result is a diff with another job, that job’s ID.
   */
  public $result_diff_job_id = 0;

  /**
   * @var array Files to save for followup comment.
   */
  private $files_to_save = [];

  /**
   * Creates a new entity.
   *
   * @see entity_create()
   */
  public function __construct(array $values = array()) {
    try {
      parent::__construct($values, 'pift_ci_job');
    }
    catch (Exception $e) {
      // Thats a really terrible use of an exception.
    }

    if (is_null($this->uid)) {
      $this->uid = $GLOBALS['user']->uid;
    }
  }

  /**
   * Load an PiftCiJob entity.
   *
   * @param int $job_id ID of the CI job to load.
   */
  public static function load($job_id) {
    return entity_load_single('pift_ci_job', $job_id);
  }

  /**
   * Save a pift_ci_job entity.
   *
   * @param bool $complete
   *   TRUE to update the issue & send email notifications as needed.
   * @param bool $preserve_timestamp
   *   TRUE to keep the existing timestamp of the job.
   *
   * @return
   * @throws \Exception
   */
  public function save($complete = FALSE, $preserve_timestamp = FALSE) {
    if (!$preserve_timestamp) {
      $this->updated = REQUEST_TIME;
    }
    // New test.
    if (empty($this->created)) {
      $this->created = REQUEST_TIME;

      if ($this->getProject()->type !== 'project_core') {
        if (empty($this->core_branch)) {
          $this->core_branch = '~development';
        }
        $this->core_branch = pift_get_current_core($this->core_branch, $this->getReleaseCategory());
      }
    }

    // Check that the job is going from running to not running.
    if ($complete && !$this->isRunning() && entity_load_unchanged('pift_ci_job', $this->job_id)->isRunning()) {
      // Check for issue updates - test for a file did not pass and did not
      // error.
      // todo PIFT_CI_TYPE_MERGE_REQUEST too?
      if ($this->target_type === PIFT_CI_TYPE_FILE && $this->result !== PIFT_CI_RESULT_PASS && $this->status !== PIFT_CI_STATUS_ERROR) {
        // And is in an envionment & release with issue testing.
        if ($this->getCurrentTrigger() === PIFT_CI_TEST_ISSUE) {
          $node = node_load($this->issue_nid);
          $file_items = field_get_items('node', $node, 'field_issue_files');
          // And the issue meets the criteria for an auto followup.
          if (pift_test_check_criteria_issue($node) && !empty($file_items)) {
            $file = (object) $this->getFile();
            $file_changes = nodechanges_get_changed_cids_keyed($node, 'field_issue_files', 'fid');
            $cid = isset($file_changes[$file->fid]) ? $file_changes[$file->fid] : 0;

            // Only change the issue status if this is the last testable file in the
            // field_issue_files stack.
            foreach (array_reverse($file_items) as $file_item) {
              // Is this a testable file?
              if (!pift_test_check_criteria_file((object) $file_item)) {
                // Not a testable file, check next entry.
                continue;
              }
              // Does this file match the file triggering the followup?
              if ($file_item['uri'] == $file->uri) {
                $node->field_issue_status[LANGUAGE_NONE][0]['value'] = PIFT_FOLLOWUP_FAIL;
              }
              break;
            }

            $account = user_load(variable_get('project_issue_followup_user', drupal_anonymous_user()));
            $node->nodechanges_uid = $account->uid;
            $node->nodechanges_comment['comment_body'][LANGUAGE_NONE][0] = [
              'value' => theme('pift_auto_followup', ['nid' => $node->nid, 'cid' => $cid, 'filename' => $file->filename, 'files_to_save' => $this->files_to_save, 'job' => $this]),
              'format' => filter_default_format($account),
            ];
            // Clear out any null file placeholder within the field_issue_files field, to
            // prevent them causing array_flip() PHP notices. See [#2061743].
            pift_sanitize_file_field($node);
            // Create the auto followup.
            node_save($node);
          }
        }
      }

      if ($this->target_type === PIFT_CI_TYPE_BRANCH) {
        // Unblock waiting tests if possible.
        if ($this->result === PIFT_CI_RESULT_PASS) {
          $query = new EntityFieldQuery();
          $query->entityCondition('entity_type', 'pift_ci_job')
            ->propertyCondition('status', PIFT_CI_STATUS_WAITING)
            ->propertyCondition('environment', $this->environment)
            ->propertyCondition('release_nid', $this->release_nid)
            ->execute();
          $result = $query->execute();
          if (isset($result['pift_ci_job'])) {
            foreach (entity_load('pift_ci_job', array_keys($result['pift_ci_job'])) as $job) {
              $job->status = PIFT_CI_STATUS_QUEUE;
              $job->save();
            }
          }
        }

        // Clear the download table cache.
        drupal_register_shutdown_function('project_release_download_table', $this->getProject()->nid, TRUE);
      }

      // Load notifications.
      $notifications = array();
      $job_specific = array();
      foreach ($this->getNotifications() as $notification) {
        if (!empty($notification->job_id)) {
          $job_specific[] = $notification->notification_id;
        }
        // De-duplicate notifications, prefering notifications for all over
        // fail-only.
        if (!isset($notifications[$notification->mail]) || $notifications[$notification->mail] !== PIFT_CI_SUBSCRIBE_ALL) {
          $notifications[$notification->mail] = $notification->notify_on;
        }
      }

      // Send notifications.
      foreach ($notifications as $mail => $notify_on) {
        if ($notify_on === PIFT_CI_SUBSCRIBE_NONPASS && $this->result === PIFT_CI_RESULT_PASS) {
          continue;
        }
        drupal_mail('pift', 'ci_notification', $mail, language_default(), array('job' => $this));
      }

      // Clean up job-specific notifications.
      if (!empty($job_specific)) {
        db_delete('pift_ci_notification')
          ->condition('notification_id', $job_specific)
          ->execute();
      }
    }

    return parent::save();
  }

  /**
   * Check if a test is actively running.
   */
  public function isRunning() {
    return $this->status === PIFT_CI_STATUS_QUEUE || $this->status === PIFT_CI_STATUS_SENT || $this->status === PIFT_CI_STATUS_RUNNING;
  }

  /**
   * Check if the current user has access to edit the job.
   */
  public function editAccess() {
    // Check if the user created the job.
    if (user_is_logged_in() && $GLOBALS['user']->uid == $this->uid) {
      return TRUE;
    }

    // Check access for the project.
    return pift_pages_project_testing_settings_access($this->getProject());
  }

  /**
   * Get the project node for this test's release branch.
   */
  public function getProject() {
    return node_load(project_release_get_release_project_nid(node_load($this->release_nid)));
  }

  /**
   * Get the release category for the test's release node.
   */
  public function getReleaseCategory() {
    if (!isset($this->release_category)) {
      $release_wrapper = entity_metadata_wrapper('node', node_load($this->release_nid));
      $this->release_category = $release_wrapper->field_release_category->value();
    }

    return $this->release_category;
  }

  /**
   * Get the file object for the test.
   */
  public function getFile() {
    if (!isset($this->file)) {
      $issue_node = node_load($this->issue_nid);
      foreach ($issue_node->field_issue_files[LANGUAGE_NONE] as $file) {
        if ($file['fid'] == $this->file_id) {
          $this->file = $file;
          break;
        }
      }
    }
    return $this->file;
  }

  /**
   * Get a test's title for humans.
   */
  public function getTitle() {
    $environment = $this->getEnvironment();
    $environment_title = $environment['title'];
    if (!empty($this->core_branch)) {
      $environment_title .= ', Drupal ' . check_plain($this->core_branch);
    }

    // Load project* nodes and set breadcrumbs.
    if (!empty($this->file_id)) {
      $file = $this->getFile();
      return t('<code>@file</code> test with @environment', [
        '@file' => drupal_basename($file['uri']),
        '@environment' => $environment_title,
      ]);
    }
    elseif ($this->target_type === PIFT_CI_TYPE_MERGE_REQUEST) {
      $merge_request = $this->getMergeRequestInfo();
      return t('Merge request !merge_request test with @environment', [
        '!merge_request' => l('!' . $merge_request['iid'], $merge_request['issue_fork']->mergeRequestUrl($merge_request['iid']), ['external' => TRUE]),
        '@environment' => $environment_title,
      ]);
    }
    else {
      return t('@branch test with @environment', [
        '@branch' => node_load($this->release_nid)->field_release_version[LANGUAGE_NONE][0]['value'],
        '@environment' => $environment_title,
      ]);
    }
  }

  /**
   * Get a test’s target’s title for humans.
   *
   * @return string
   *   The human-readable title of the test target.
   */
  public function getTargetTitle() {
    switch ($this->target_type) {
      case PIFT_CI_TYPE_BRANCH:
        return node_load($this->release_nid)->field_release_version[LANGUAGE_NONE][0]['value'];

      case PIFT_CI_TYPE_FILE:
        return drupal_basename($this->getFile()['uri']);

      case PIFT_CI_TYPE_MERGE_REQUEST:
        return t('merge request !@iid', ['@iid' => $this->getMergeRequestInfo()['iid']]);
    }
  }

  /**
   * Get the environment configuration for this job.
   */
  public function getEnvironment() {
    $environments = variable_get('pift_ci_environments');
    return $environments[$this->environment];
  }

  /**
   * When target_type is PIFT_CI_TYPE_MERGE_REQUEST, get the issue fork & IID.
   *
   * @return array
   *   - issue_fork: a DrupalorgIssueFork entity.
   *   - iid: the merge request IID.
   */
  public function getMergeRequestInfo() {
    $result = (new EntityFieldQuery())->entityCondition('entity_type', 'drupalorg_issue_fork')
      ->propertyCondition('nid', $this->issue_nid)
      ->propertyCondition('project_nid', entity_metadata_wrapper('node', node_load($this->release_nid))->field_release_project->raw())
      ->execute();
    $issue_fork = entity_load_single('drupalorg_issue_fork', array_keys($result['drupalorg_issue_fork'])[0]);
    return [
      'issue_fork' => $issue_fork,
      'iid' => $issue_fork->getMergeRequestIID($this->issue_branch, $this->merge_target),
    ];
  }

  /**
   * Get the project configuration for this job.
   */
  public function getProjectInfo() {
    return pift_ci_project_info(project_release_get_release_project_nid(node_load($this->release_nid)));
  }

  /**
   * Get currently-configured testing trigger, if any.
   */
  public function getCurrentTrigger() {
    if ($this->custom) {
      return NULL;
    }

    $project_info = $this->getProjectInfo();
    if (isset($project_info[$this->environment][$this->release_nid])) {
      foreach ($project_info[$this->environment][$this->release_nid] as $core_branch => $testing_trigger) {
        // Dereference the core branch label to compare to what the test ran
        // with.
        if (pift_get_current_core($core_branch, $this->getReleaseCategory()) === $this->core_branch) {
          return $testing_trigger;
        }
      }
    }

    return NULL;
  }

  /**
   * Check if a test should be blocked on branch passing.
   */
  public function checkBranchBlock() {
    // Only queued file tests can be blocked.
    if ($this->status !== PIFT_CI_STATUS_QUEUE || !in_array($this->target_type, [PIFT_CI_TYPE_FILE, PIFT_CI_TYPE_MERGE_REQUEST])) {
      return FALSE;
    }

    // Bypassing has been requested.
    if (!empty($this->bypass_uid)) {
      return FALSE;
    }

    // Only block if the environment & release is set to issue testing.
    if ($this->getCurrentTrigger() !== PIFT_CI_TEST_ISSUE) {
      return FALSE;
    }

    // Block if the branch has not passed.
    if ($branch_job = $this->getCurrentBranchJob()) {
      return $branch_job->result !== PIFT_CI_RESULT_PASS;
    }

    // The branch is missing, do not block.
    return FALSE;
  }

  /**
   * Get branch job at creation time with the issue job's environment & release.
   */
  public function getLastBranchJob() {
    $query = new EntityFieldQuery();
    $result = $query->entityCondition('entity_type', 'pift_ci_job')
      ->propertyCondition('target_type', PIFT_CI_TYPE_BRANCH)
      ->propertyCondition('status', [PIFT_CI_STATUS_COMPLETE, PIFT_CI_STATUS_SWEPT])
      ->propertyCondition('environment', $this->environment)
      ->propertyCondition('release_nid', $this->release_nid)
      ->propertyCondition('updated', $this->created, '<')
      ->propertyCondition('custom', FALSE)
      ->propertyOrderBy('job_id', 'DESC')
      ->range(0, 1)
      ->execute();
    if (isset($result['pift_ci_job'])) {
      return PiftCiJob::load(reset($result['pift_ci_job'])->job_id);
    }
  }

  /**
   * Get most recent branch job with the issue job's environment & release.
   */
  public function getCurrentBranchJob() {
    $query = new EntityFieldQuery();
    $result = $query->entityCondition('entity_type', 'pift_ci_job')
      ->propertyCondition('target_type', PIFT_CI_TYPE_BRANCH)
      ->propertyCondition('status', [PIFT_CI_STATUS_COMPLETE, PIFT_CI_STATUS_SWEPT])
      ->propertyCondition('environment', $this->environment)
      ->propertyCondition('release_nid', $this->release_nid)
      ->propertyCondition('custom', FALSE)
      ->propertyOrderBy('job_id', 'DESC')
      ->range(0, 1)
      ->execute();
    if (isset($result['pift_ci_job'])) {
      return PiftCiJob::load(reset($result['pift_ci_job'])->job_id);
    }
  }

  /**
   * Takes a queued request and pushes it upstream to the api server.
   *
   * @return bool
   */
  public function requestTest() {

    // * support for profiles/drush extensions/libraries(?)

    // Get all environment definitions from settings.php
    $environments = variable_get('pift_ci_environments');
    // Get the jenkins job url for this environment

    // The api token for remote builds is stored in the environment config
    $query = $environments[$this->environment]['query'];

    $release = node_load($this->release_nid);
    $release_wrapper = entity_metadata_wrapper('node', $release);
    $category = $this->getReleaseCategory();
    $project_type = $release_wrapper->field_release_project->value()->type;

    if (!in_array($project_type, ['project_core','project_module','project_theme'])) {
        watchdog('pift_ci', 'Unsupported project testing type.');
        return FALSE;
    }

    // TODO: for now, we'll gather the job yml template definition based off of
    // whether it's d7 or d8.
    // Eventually we'll want to figure out which template to gather based off of
    // the trigger assignment to the job, and set the assessment_target
    // accordingly.
    $assessment_target = 'development';
    if ($category === 'legacy') {
      // All D7 testing.
      $build_template = $this->getJobTemplate('legacydevelopment');
      $legacy_extension = '_d7';
    } else {
      $build_template = $this->getJobTemplate('development');
      $legacy_extension = '';
    }

    // Set the project name and type
    $project_name = $release_wrapper->field_release_project->field_project_machine_name->value();
    $build_template['build']['codebase']['project-name'] = $project_name;
    $build_template['build']['codebase']['project-type'] = str_replace('project_', '', $project_type);
    $build_template['build']['environment']['startcontainers']['runcontainers']['phpversion'] = $query['DCI_PHPVersion'];

    // TODO: Remove this when php8 can composer install. see https://www.drupal.org/project/project_issue_file_test/issues/3156521
    if ($this->environment === 'php8.0_mysql5.7_ignore') {
      $build_template['build']['codebase']['assemble_codebase']['composer.core_install']['options'] = 'install --prefer-dist --no-suggest --no-interaction --ignore-platform-reqs';
    }

    // TODO: Remove this when phpspec/prophecy works on 8.2 see https://www.drupal.org/project/project_issue_file_test/issues/3318421
    if ($this->environment === 'php8.2_mysql8_ignore') {
      $build_template['build']['codebase']['assemble_codebase']['composer.core_install']['options'] = 'install --prefer-dist --no-suggest --no-interaction --ignore-platform-req=php';
      $build_template['build']['codebase']['assemble_codebase']['gather_dependencies']['options'] = '--prefer-stable --no-progress --no-suggest --no-interaction --ignore-platform-req=php';
    }

    if(strpos($this->environment, 'unlocked') > 0){
      $build_template['build']['codebase']['assemble_codebase']['composer.core_update']['options'] = 'update --with-all-dependencies --with phpunit/phpunit:$(composer show -i |grep phpunit\/phpunit |awk \'{print $2}\')';
      $build_template['build']['codebase']['assemble_codebase']['composer.core_install_again']['options'] = 'install';
    }

    // TODO: change DCI_DBVersion into DCI_DBtype and DCI_DB_VERSION that are already split
    $build_template['build']['environment']['db-type'] = explode('-', $query['DCI_DBVersion'], 2)[0];
    $build_template['build']['environment']['db-version'] = explode('-', $query['DCI_DBVersion'], 2)[1];
    unset($query['DCI_DBVersion'],$query['DCI_PHPVersion']);

    // Check if the branch is blocking this test.
    if ($this->checkBranchBlock()) {
      $this->status = PIFT_CI_STATUS_WAITING;
      $this->message = t('Waiting for branch to pass');
      $this->save();
      return FALSE;
    }

    if ($this->target_type === PIFT_CI_TYPE_FILE) {
      // Gather the patch for this test and add it to the build config.
      $file = file_load($this->file_id);
      $patch_url = file_create_url($file->uri);

      $build_template['build']['codebase']['assemble_codebase']['fetch']['files'][] = ['from' => $patch_url, 'to' => '.'];
      if ($project_type === 'project_core') {
        $build_template['build']['codebase']['assemble_codebase']['patch']['patches'][] = [
          'from' => drupal_basename($patch_url),
          'to' => '.'
        ];
      } else {
        $build_template['build']['codebase']['assemble_codebase']['patch']['patches'][] = [
          'from' => drupal_basename($patch_url),
          'to' => 'PROJECTVCSREPO_DIR'
        ];
      }
    }

    $project_nid = $release_wrapper->field_release_project->nid->value();

    $repository = variable_get('pift_git_base_url', 'git://git.drupal.org/project/') . $release_wrapper->field_release_project->value()->versioncontrol_project['repo']->name . '.git';
    if ($this->target_type === PIFT_CI_TYPE_MERGE_REQUEST) {
      $merge_request = $this->getMergeRequestInfo();
      if (!$merge_request['issue_fork']->mergeRequestCanBeMerged($merge_request['iid'])) {
        $this->status = PIFT_CI_STATUS_COMPLETE;
        $this->result = PIFT_CI_RESULT_FAIL;
        $this->message = t('Not currently mergeable.');
        $this->save(TRUE);
        return FALSE;
      }

      $branch = 'merge-request-' . $merge_request['iid'];
      $reftype = 'merge-request';
      $build_template['build']['codebase']['target-branch'] = $this->merge_target;

    }
    else {
      // Not a merge request test.
      $build_template['build']['codebase']['commit-hash'] = $this->commit_id;

      if (empty($this->issue_branch)) {
        $branch = $release->field_release_vcs_label[LANGUAGE_NONE][0]['value'];
        $reftype = 'branch';
      }
      else {
        $branch = $this->issue_branch;
        $reftype = 'feature-branch';
      }
    }
    $composer_namespace = project_composer_get_project_namespace($project_nid, $release_wrapper->field_release_category->value());

    $build_template['build']['codebase']['composer-namespace'] = $composer_namespace;
    $build_template['build']['codebase']['reference-type'] = $reftype;
    $build_template['build']['codebase']['repository'] = $repository;
    $build_template['build']['codebase']['reference'] = $branch;

    if ($project_type === 'project_core') {
      // Core test.
      $core_repo = $repository;
      $core_branch = $branch;
    }
    else {
      // Contrib test.
      $core_repo = variable_get('pift_git_base_url', 'git://git.drupal.org/project/') . node_load(variable_get('pift_pid', 3060))->versioncontrol_project['repo']->name . '.git';
      $core_branch = $this->core_branch;
      $repository_data = [
        'repo' => $repository,
        'branch' => $branch,
      ];
      if ($project_type !== 'project_core' && $this->target_type === PIFT_CI_TYPE_MERGE_REQUEST) {
        $repository_data['merge-request'] = $merge_request['iid'];
      }
      else {
        $repository_data['commit-hash'] = $this->commit_id;
      }
      $build_template['build']['codebase']['assemble_codebase']["checkout.contrib"]['repositories'][] = $repository_data;
    }
    $repository_data = ['repo' => $core_repo];
    if ($project_type === 'project_core' && $this->target_type === PIFT_CI_TYPE_MERGE_REQUEST) {
      $repository_data['merge-request'] = $merge_request['iid'];
    }
    else {
      $repository_data['branch'] = $core_branch;
    }
    $build_template['build']['codebase']['assemble_codebase']["checkout_core{$legacy_extension}"]['repositories'][] = $repository_data;

    $project_repo = versioncontrol_project_repository_load($project_nid);
    // This is where we load in a build.yml if we've got one, and replace
    // the assessment section.
    // There should be only one config file in the db for the combination of
    // path + repo_id + branch
    if ($project_nid == DRUPALORG_CORE_NID){
      $filepath = '/core/drupalci.yml';
    } else {
      $filepath = '/drupalci.yml';
    }
    $config_query = db_select('pift_ci_config', 'c')
      ->condition('c.file_path', $filepath)
      ->condition('c.file_type', PIFT_CI_CONFIG_TYPE_SYSTEM)
      ->condition('c.repo_id', $project_repo->repo_id)
      ->condition('c.branch', $release->field_release_vcs_label[LANGUAGE_NONE][0]['value'])
      ->fields('c', ['contents']);
    $config_file = $config_query->execute()->fetchAssoc();

    if (!empty($config_file)) {
      try {
        $drupalci_config = $this->getYamlParser()->parse($config_file['contents']);
        $build_key = "build.${assessment_target}";
        if (isset($drupalci_config[$build_key])) {
          $build_template['build']['assessment'] = $drupalci_config[$build_key]['assessment'];
        }
        else if (isset($drupalci_config['build'])) {
          $build_template['build']['assessment'] = $drupalci_config['build']['assessment'];
        }
        else {
          watchdog('pift_ci', 'Invalid drupalci.yml for @project', ['@project' => $project_name], WATCHDOG_WARNING);
        }
      }
      catch (Symfony\Component\Yaml\Exception\ParseException $e) {
        watchdog('pift_ci', 'Invalid drupalci.yml for @project', ['@project' => $project_name], WATCHDOG_WARNING);
      }
    }
    $jenkins_job_url = $this->getJenkinsBuildUrl($category, $project_type);
    $buildyml = $this->getYamlParser()->dump($build_template, PHP_INT_MAX, 2, 0);
    try {
      $response = pift_get_jenkins_client()->post("{$jenkins_job_url}/build", [
        'query' => $query,
        'multipart' => [
          [
            'name' => 'json',
            'contents' => json_encode(['parameter' => [
              [ 'name' => 'builds/build.yml', 'file' => 'file0'],
              [ 'name' => 'Drupal_JobID', 'value' => $GLOBALS['base_url'] . ':' . $this->job_id],
              [ 'name' => 'DCI_PHPVersion', 'value' => $build_template['build']['environment']['startcontainers']['runcontainers']['phpversion']],
            ]
            ]),
            'headers' => ['Content-Type' => 'application/json']
          ],
          [
            'name' => 'file0',
            'contents' => $buildyml,
            'filename' => 'build.yml',
          ],
        ],
      ]);
      $this->status = PIFT_CI_STATUS_SENT;
      $this->message = t('Queued');
      $this->save();

      return TRUE;
    }
    catch (RequestException $e) {
      watchdog('pift_ci', 'CI API submission failure: @error; @data', [
        '@error' => $e->getMessage(),
        '@data' => ($response = $e->getResponse()) ? Psr7\str($response) : '',
      ], WATCHDOG_WARNING);
      return FALSE;
    }
  }

  public function requestResult() {
    if (!lock_acquire('PiftCiJob::requestResult:' . $this->job_id)) {
      watchdog('pift_ci', 'Skipping fetching results for @job_id, already locked', ['@job_id' => $this->job_id]);
      return;
    }

    watchdog('pift_ci', 'Fetching CI results for @job_id', ['@job_id' => $this->job_id], WATCHDOG_INFO);

    try {
      $job_response = pift_get_jenkins_client()->get($this->ci_url . '/api/json', ['timeout' => 60.0]);
    } catch (RequestException $e) {
      if ($e->getCode() == 404) {
        $this->status = PIFT_CI_STATUS_ERROR;
        $this->message = t('CI job missing');
        $this->save(TRUE);
        return;
      } else {
        watchdog('pift_ci', 'CI results fetch failure: @error', [
          '@error' => $e->getMessage()
        ], WATCHDOG_WARNING);
        return;
      }
    }

    $job_data = json_decode($job_response->getBody()->getContents());

    if ($job_data->building) {
      // Job is not complete.
      return;
    }

    // We'll assume that something went wrong until we get the buildoutcome
    // data that informs us otherwise.
    $this->message = t('CI error');

    // Get build outcome data.
    $project_directory = '';
    $checkstyle_results = [];
    $filenames_to_save = [
      'codesniffer_autofix.patch' => t('Patch with automated coding standards fixes.'),
      'codesniffer_fixes.patch' => t('Interdiff of automated coding standards fixes only.'),
    ];
    foreach ($job_data->artifacts as $artifact) {
      if ($artifact->fileName == 'buildoutcome.json') {
        try {
          $build_response = pift_get_jenkins_client()->get($this->ci_url . '/artifact/' . $artifact->relativePath, ['timeout' => 60.0]);
          $build_data = json_decode($build_response->getBody()->getContents());
          $this->message = $build_data->buildLabel;
          $this->build_details = $build_data->buildDetails;
        } catch (RequestException $e) {
          watchdog('pift_ci', 'CI buildoutcome unavailable fetch failure: @error', ['@error' => $e->getMessage()], WATCHDOG_WARNING);
          $this->build_details = '';
        }
      }
      elseif ($artifact->fileName === 'project_directory.json') {
        try {
          $build_response = pift_get_jenkins_client()->get($this->ci_url . '/artifact/' . $artifact->relativePath, ['timeout' => 60.0]);
          $project_directory = json_decode($build_response->getBody()->getContents())->projectDirectory;

        } catch (RequestException $e) {
          watchdog('pift_ci', 'CI project_directory unavailable fetch failure: @error', ['@error' => $e->getMessage()], WATCHDOG_WARNING);
        }
      }
      elseif ($artifact->fileName === 'checkstyle.xml') {
        $checkstyle_results[] = $this->ci_url . '/artifact/' . $artifact->relativePath;
      }
      elseif (isset($filenames_to_save[$artifact->fileName])) {
        // Check if file has data.
        try {
          $build_response = pift_get_jenkins_client()->head($this->ci_url . '/artifact/' . $artifact->relativePath, [
            'timeout' => 60.0]);
          if ($build_response->getHeader('Content-Length')[0] > 1) {
            $this->files_to_save[$this->ci_url . '/artifact/' . $artifact->relativePath] = $filenames_to_save[$artifact->fileName];
          }
        } catch (RequestException $e) {
          watchdog('pift_ci', 'CI @artifact empty fetch failure: @error', ['@error' => $e->getMessage()], WATCHDOG_WARNING);
        }
      }
    }

    // Get the checkstyle results regardles of the build state.
    $this->requestCheckstyleResult($project_directory, $checkstyle_results);

    // Check for general failure.
    if ($job_data->result === 'FAILURE') {
      if (isset($build_data)) {
        // Build outcome has an explanation of the failure.
        $this->status = PIFT_CI_STATUS_ERROR;
        $this->save(TRUE);
      }
      else {
        // No build outcome found, something went terribly wrong.
        $this->status = PIFT_CI_STATUS_ERROR;
        $this->message = t('CI error');
        $this->save();
      }
      return;
    }
    if ($job_data->result === 'ABORTED') {
      $this->status = PIFT_CI_STATUS_ERROR;
      $this->message = t('CI aborted');
      $this->save(TRUE);
      return;
    }

    try {
      $test_response_body = pift_get_jenkins_client()->get($this->ci_url . '/testReport/api/json')->getBody();
      $test_response = '';
      $pift_max_result_size = variable_get('pift_max_result_size', 100*1024*1024);
      while (!$test_response_body->eof() && strlen($test_response) < $pift_max_result_size) {
        $test_response .= $test_response_body->read($pift_max_result_size - strlen($test_response));
      }
      if (!$test_response_body->eof()) {
        $this->status = PIFT_CI_STATUS_ERROR;
        $this->message = t('Test results are too large!');
        $this->save(TRUE);
        return;
      }
      $test_data = json_decode($test_response);
      if (is_null($test_data)) {
        watchdog('pift_ci', 'CI test results JSON parse failure, error @code', ['@code' => json_last_error()], WATCHDOG_WARNING);
        return;
      }

      // Save result data.
      db_query('DELETE FROM {pift_ci_job_result} WHERE job_id = :job_id', array(':job_id' => $this->job_id));
      db_query('DELETE FROM {pift_ci_job_result_diff} WHERE job_id = :job_id', array(':job_id' => $this->job_id));
      $job_result = [];
      foreach ($test_data->suites as $suite) {
        foreach ($suite->cases as $case) {
          $suite_name = preg_replace('|^TODO_ Set Package\.|', '', $suite->name);
          $class_name = preg_replace('|^' . preg_quote($suite_name . '.', '|') . '|', '', $case->className);
          $test_name = truncate_utf8($case->name, 128);
          $job_result[$suite_name . ':' . $class_name . ':' . $test_name][] = [
            'job_id' => $this->job_id,
            'suite' => $suite_name,
            'class' => $class_name,
            'test' => $test_name,
            'result' => ($case->status === 'PASSED' || $case->status === 'FIXED') ? PIFT_CI_RESULT_PASS : PIFT_CI_RESULT_FAIL,
            'error' => $case->errorDetails,
            'output' => $case->stdout,
          ];
        }
      }
      // Check if the results can be stored as a diff.
      if (in_array($this->target_type, [PIFT_CI_TYPE_FILE, PIFT_CI_TYPE_MERGE_REQUEST]) && ($branch_job = $this->getLastBranchJob())) {
        $this->result_diff_job_id = $branch_job->job_id;
        $branch_job_result = $branch_job->getResultPackedArray();

        // job_result_id for suite/class/test present in branch, but not issue.
        if ($diff = array_map('array_keys', array_values(array_diff_key($branch_job_result, $job_result)))) {
          $query = db_insert('pift_ci_job_result_diff')
            ->fields(['job_id', 'job_result_id']);
          foreach (call_user_func_array('array_merge', $diff) as $job_result_id) {
            $query->values(['job_id' => $this->job_id, 'job_result_id' => $job_result_id]);
          }
          $query->execute();
        }

        // Present in both.
        $remove_from_issue = [];
        foreach (array_intersect(array_keys($job_result), array_keys($branch_job_result)) as $key) {
          // Results are same in both.
          $packed = [];
          foreach ($job_result[$key] as $test) {
            $packed[] = $test['result'] . ':' . $test['error'] . ':' . $test['output'];
          }
          if (count(array_intersect($packed, $branch_job_result[$key])) === count($branch_job_result[$key])) {
            unset($job_result[$key]);
          }
        }
      }
      // Write results.
      foreach ($job_result as $test) {
        foreach ($test as $result) {
          drupal_write_record('pift_ci_job_result', $result);
        }
      }

      // Save job data.
      if ($test_data->passCount == 0 && $test_data->failCount == 0 && $test_data->skipCount == 0) {
        $this->message = "No tests found";
      }
      else {
        $message = array();
        foreach (array('passCount' => '@count pass', 'failCount' => '@count fail', 'skipCount' => '@count skip') as $key => $text) {
          if ($test_data->$key) {
            $message[] = t($text, array('@count' => number_format($test_data->$key)));
          }
        }
        $this->message = implode(', ', $message);
      }
      $this->result = $test_data->failCount ? PIFT_CI_RESULT_FAIL : PIFT_CI_RESULT_PASS;
    }
    catch (Exception $e) {
      watchdog('pift_ci', 'Results fetch failure: @error', ['@error' => $e->getMessage()], WATCHDOG_WARNING, l($this->job_id, 'pift-ci-job/' . $this->job_id));
    }
    $this->status = PIFT_CI_STATUS_COMPLETE;
    $this->save(TRUE);

    lock_release('PiftCiJob::requestResult:' . $this->job_id);
  }

  /**
   * Request cancellation of a test.
   *
   * @return bool
   *   Whether the cancellation request could be made.
   */
  public function stopDispatcherJob() {
    if (empty($this->ci_url)) {
      // todo: implement cancelling queued jobs.
      return FALSE;
    }

    try {
      pift_get_jenkins_client()->post($this->ci_url . '/stop');
    } catch (RequestException $e) {
      watchdog('pift_ci', 'Job stop request failure: @error', ['@error' => $e->getMessage()], WATCHDOG_WARNING);
      return FALSE;
    }

    // Real update will come next time the job status is updated.
    $this->message = t('Cancellation requested');
    $this->save();
    return TRUE;
  }

  /**
   * Deletes a job and all of its build artifacts off of dispatcher,
   * and removes the ci_url from the job.
   *
   */
  public function purgeDispatcherJob() {
    try {
      $response = pift_get_jenkins_client()->post($this->ci_url . '/doDelete');
    } catch (RequestException $e) {
      watchdog('pift_ci', 'Job deletion failure: @error', ['@error' => $e->getMessage()], WATCHDOG_WARNING);
    } finally {
      $this->ci_url = '';
      $this->save(FALSE,TRUE);
    }
  }

  /**
   * Fetch checkstyle results.
   */
  public function requestCheckstyleResult($project_directory, $checkstyle_results) {
    db_delete('pift_ci_job_checkstyle')->condition('job_id', $this->job_id)->execute();
    $this->checkstyle = 0;

    $project_directory_pattern = '|^' . preg_quote(preg_replace('|/$|', '', $project_directory), '|') . '/|';
    foreach ($checkstyle_results as $checkstyle_result) {
      try {
        $style_response = pift_get_jenkins_client()->get($checkstyle_result, ['timeout' => 60.0]);
      } catch (RequestException $e) {
        watchdog('pift_ci', 'CI checkstyle results fetch failure: @error', ['@error' => $e->getMessage()], WATCHDOG_WARNING);
        continue;
      }

      if (($response = $style_response->getBody()->getContents()) && $response[0] === '<' && ($data = simplexml_load_string($response)) !== FALSE) {
        foreach ($data->file as $file) {
          $filename = truncate_utf8(preg_replace($project_directory_pattern, '', $file['name']), 255, FALSE, TRUE);
          foreach ($file->error as $error) {
            $row = [
              'job_id' => $this->job_id,
              'file' => $filename,
              'line' => $error['line'],
              'message' => truncate_utf8($error['message'], 255, FALSE, TRUE),
            ];
            drupal_write_record('pift_ci_job_checkstyle', $row);
            $this->checkstyle += 1;
          }
        }
      }
    }
  }

  /**
   * Find notifications to be sent for this job.
   */
  public function getNotifications($account = NULL) {
    $query = db_select('pift_ci_notification', 'n')
      ->fields('n', array('notification_id', 'uid', 'notify_on', 'own', 'job_id', 'environment', 'release_nid', 'file_id'));
    $query->join('users', 'u', 'u.uid = n.uid');
    $query->fields('u', array('mail'));

    if (!is_null($account)) {
      $query->condition('n.uid', $account->uid);
    }

    $condition = db_or()
      ->condition('n.job_id', $this->job_id);
    if ($this->target_type === PIFT_CI_TYPE_BRANCH) {
      $condition->condition(db_and()
        ->condition('n.release_nid', $this->release_nid)
        ->condition('n.environment', $this->environment)
      );
    }
    $query->condition($condition);

    return $query->execute()->fetchAll();
  }

  /**
   * Get all results packed in a structure for diffing.
   */
  public function getResultPackedArray() {
    $return = [];

    $result = db_query('SELECT job_result_id, suite, class, test, result, error, output FROM {pift_ci_job_result} WHERE job_id = :job_id ORDER BY job_result_id', [':job_id' => $this->job_id]);
    foreach ($result as $job_result) {
      $return[$job_result->suite . ':' . $job_result->class . ':' . $job_result->test][$job_result->job_result_id] = $job_result->result . ':' . $job_result->error . ':' . $job_result->output;
    }

    return $return;
  }

  /**
   * Get all results in a structure for display.
   *
   * @param array $result_condition
   *   Optional result types to filter on.
   */
  public function getResultArray(array $result_condition = []) {
    $return = [];
    $query = db_select('pift_ci_job_result', 'r')
      ->condition('r.job_id', $this->job_id)
      ->fields('r', ['suite', 'class', 'test', 'result', 'error', 'output'])
      ->orderBy('r.job_result_id');
    if ($result_condition) {
      $query->condition('r.result', $result_condition);
    }
    $result = $query->execute();
    foreach ($result as $job_result) {
      $return[$job_result->suite][$job_result->class][$job_result->test] = $job_result;
    }
    if (!empty($this->result_diff_job_id)) {
      // Add in results from diff job.
      $query = db_select('pift_ci_job_result', 'r')
        ->condition('r.job_id', $this->result_diff_job_id)
        ->fields('r', ['suite', 'class', 'test', 'result', 'error', 'output'])
        ->orderBy('r.job_result_id');
      $query->leftJoin('pift_ci_job_result_diff', 'd', 'd.job_result_id = r.job_result_id AND d.job_id = :job_id', [':job_id' => $this->job_id]);
      $query->condition('d.job_result_id', NULL, 'IS NULL');
      if ($result_condition) {
        $query->condition('r.result', $result_condition);
      }
      $result = $query->execute();
      foreach ($result as $job_result) {
        if (!isset($return[$job_result->suite][$job_result->class][$job_result->test])) {
          $return[$job_result->suite][$job_result->class][$job_result->test] = $job_result;
        }
      }
    }

    return $return;
  }

  /**
   *  Creates a Yaml parser that the job can use for reading in templates
   *  and for dumping the build into yaml to send.
   */
  protected function getYamlParser() {
    static $yaml_parser;

    if (is_null($yaml_parser)) {
      $yaml_parser = new Yaml();
    }

    return $yaml_parser;
  }

  /**
   * Determines which URL to use when sending to Jenkins.
   *
   * - Is it d7 or d8, ($category)
   * - Is it contrib or core ($project_type)
   * - Is it an issue-related or branch test ($this->target_type)
   *
   * Todo: make project_type part of the entity
   *
   * @param $category
   * @param $project_type
   *
   * @return mixed
   */
  protected function getJenkinsBuildUrl($category, $project_type) {
    $environments = variable_get('pift_ci_environments');
    // Set a default first.
    $jenkins_job_url = $environments[$this->environment]['url'];
    if ($category === 'legacy') {
      // All D7 testing.
      $jenkins_job_url = $environments[$this->environment]['url_d7'];
    }
    else {
      // D8 test
      if ($project_type === 'project_core') {
        // D8 Core testing
        if (in_array($this->target_type, [PIFT_CI_TYPE_FILE, PIFT_CI_TYPE_MERGE_REQUEST])) {
          // D8 Core Patch Testing
          if (isset($environments[$this->environment]['url_file'])) {
            $jenkins_job_url = $environments[$this->environment]['url_file'];
          }
        }
        else {
          $jenkins_job_url = $environments[$this->environment]['url'];
          // D8 Core Branch Testing
        }
      }
      else {
        // Contrib Testing
        if (in_array($this->target_type, [PIFT_CI_TYPE_FILE, PIFT_CI_TYPE_MERGE_REQUEST])) {
          // D8 Contrib Patch Testing
          if (isset($environments[$this->environment]['url_contrib_file'])) {
            $jenkins_job_url = $environments[$this->environment]['url_contrib_file'];
          }
        }
        else {
          // Use the environment’s url_contrib if it exists.
          // D8 Contrib Branch Testing
          if (isset($environments[$this->environment]['url_contrib'])) {
            $jenkins_job_url = $environments[$this->environment]['url_contrib'];
          }
        }
      }
    }
    return $jenkins_job_url;
  }

  /**
   * Returns a php array of the yaml file for a particular assessment target.
   *
   * @param $target
   *
   * @return mixed
   */
  protected function getJobTemplate($target) {
    $yamlfile = file_get_contents(drupal_get_path('module', 'pift') . "/job_templates/${target}.yml");
    return $this->getYamlParser()->parse($yamlfile);
  }
}
