<?php
/**
 * @file
 * Drush integration functions for the PIFT module.
 */

/**
 * Implements hook_drush_command().
 */
function pift_drush_command() {
  $items = array();

  $items['pift-cron-test'] = array(
    'description' => 'Send DrupalCI batch for testing, and retrieve results.',
  );
  $items['pift-requeue'] = array(
    'description' => 'Re-queue queued jobs.',
  );
  $items['pift-cron-daily'] = array(
    'description' => 'Queue daily jobs.',
  );
  $items['pift-cron-fetch-results'] = array(
    'description' => 'Fetch list of DrupalCI results.',
  );
  $items['pift-cron-parse-results'] = array(
    'description' => 'Fetch DrupalCI results for each job.',
  );
  $items['pift-sweep-data'] = [
    'description' => 'Delete outdated results and data',
  ];
  $items['pift-disable-core-label'] = [
    'description' => 'Disable testing with a core label, and testing configurations using it.',
    'arguments' => [
      'label' =>  'Core label machine name, like “~development”',
    ],
  ];
  $items['pift-queue-branch-test'] = [
    'description' => 'Queue an individual branch test for a Git push.',
    'arguments' => [
      'project' => 'Project machine name.',
      'branch' => 'Branch name.',
      'commit-id' => 'Git commit hash.'
    ],
  ];

  return $items;
}

/**
 * Implements hook_drush_help().
 */
function pift_drush_help($key) {
  switch ($key) {
    case 'error:PIFT_NO_BRANCHES_CONFIGURED':
      return dt('No branches configured.');

    case 'error:PIFT_PROJECT_NOT_FOUND':
      return dt('Project not found');

    case 'error:PIFT_RELEASE_NOT_FOUND':
      return dt('Project not found');
  }
}

/**
 * Send DrupalCI batch for testing.
 */
function drush_pift_cron_test() {
  $query = new EntityFieldQuery();
  $jobs_running = $query->entityCondition('entity_type', 'pift_ci_job')
    ->propertyCondition('status', [PIFT_CI_STATUS_SENT, PIFT_CI_STATUS_RUNNING])
    ->count()->execute();
  $limit = min(variable_get('pift_ci_limit', 10), max(variable_get('pift_ci_concurrency_limit', 50) - $jobs_running, 0));
  if ($limit <= 0) {
    return;
  }

  // Request testing of queued jobs.
  $query = new EntityFieldQuery();
  $query->entityCondition('entity_type', 'pift_ci_job')
    ->propertyCondition('status', PIFT_CI_STATUS_QUEUE);
  $result = $query->execute();
  if (isset($result['pift_ci_job'])) {
    foreach (entity_load('pift_ci_job', array_keys($result['pift_ci_job'])) as $job) {
      drush_log(dt('Requesting test for !job_id', ['!job_id' => $job->job_id]));
      $limit -= $job->requestTest();
      if ($limit <= 0) {
        break;
      }
    }
  }

  // Re-request testing of jobs stuck over 12h.
  $query = new EntityFieldQuery();
  $query->entityCondition('entity_type', 'pift_ci_job')
    ->propertyCondition('status', PIFT_CI_STATUS_SENT)
    ->propertyCondition('updated', REQUEST_TIME - 60*60*12, '<');
  $result = $query->execute();
  if (isset($result['pift_ci_job'])) {
    foreach (entity_load('pift_ci_job', array_keys($result['pift_ci_job'])) as $job) {
      drush_log(dt('Re-requesting test for !job_id', ['!job_id' => $job->job_id]));
      $limit -= $job->requestTest();
      if ($limit <= 0) {
        break;
      }
    }
  }

  // Stop trying if branch has blocked for 4 weeks.
  $query = new EntityFieldQuery();
  $query->entityCondition('entity_type', 'pift_ci_job')
    ->propertyCondition('status', PIFT_CI_STATUS_WAITING)
    ->propertyCondition('updated', REQUEST_TIME - 4*7*24*60*60, '<');
  $result = $query->execute();
  if (isset($result['pift_ci_job'])) {
    foreach (entity_load('pift_ci_job', array_keys($result['pift_ci_job'])) as $job) {
      $job->message = t('Skipped, branch did not pass for 4 weeks');
      $job->status = PIFT_CI_STATUS_ERROR;
      $job->save();
    }
  }
}

/**
 * Send DrupalCI batch for testing.
 */
function drush_pift_requeue() {
  // Re-request testing of non-daily jobs that are queued. Sometimes jenkins
  // dies, and loses all queued jobs. This re-creates them all.
  $query = new EntityFieldQuery();
  $query->entityCondition('entity_type', 'pift_ci_job')
    ->propertyCondition('status', PIFT_CI_STATUS_SENT)
    ->propertyCondition('reason', 'Daily testing.%', 'NOT LIKE');
  $result = $query->execute();
  if (isset($result['pift_ci_job'])) {
    foreach (entity_load('pift_ci_job', array_keys($result['pift_ci_job'])) as $job) {
      drush_log(dt('Re-requesting previously-queued test for !job_id', ['!job_id' => $job->job_id]));
      $job->status = PIFT_CI_STATUS_QUEUE;
      $job->save();
    }
  }

  // Re-request testing of non-daily jobs that went missing in the last 3 days.
  $query = new EntityFieldQuery();
  $query->entityCondition('entity_type', 'pift_ci_job')
    ->propertyCondition('status', PIFT_CI_STATUS_ERROR)
    ->propertyCondition('message', 'CI job missing')
    ->propertyCondition('reason', 'Daily testing.%', 'NOT LIKE')
    ->propertyCondition('created', REQUEST_TIME - (60*60*24*3), '>');
  $result = $query->execute();
  if (isset($result['pift_ci_job'])) {
    foreach (entity_load('pift_ci_job', array_keys($result['pift_ci_job'])) as $job) {
      drush_log(dt('Re-requesting missing test for !job_id', ['!job_id' => $job->job_id]));
      $job->status = PIFT_CI_STATUS_QUEUE;
      $job->save();
    }
  }
}

/**
 * Fetch DrupalCI results.
 */
function drush_pift_cron_fetch_results() {
  // Get URLs of environments.
  $urls = [];
  foreach (variable_get('pift_ci_environments') as $environment) {
    $urlkeys = preg_grep('/^url/', array_keys($environment));
    foreach ($urlkeys as $key) {
      $urls[] = $environment[$key];
    }
  }

  foreach (array_unique($urls) as $url) {
    drush_log(dt('Getting results from !url', array('!url' => $url)));
    try {
      $response = pift_get_jenkins_client()->get($url . '/api/json?depth=1&tree=builds[number,actions[parameters[name,value]]]');
    }
    catch (\GuzzleHttp\Exception\RequestException $e) {
      watchdog('pift_ci', 'CI results fetch failure: @error; @data', ['@error' => $e->getMessage(), '@data' => \GuzzleHttp\Psr7\str($e->getResponse())], WATCHDOG_WARNING);
      break;
    }

    $data = json_decode($response->getBody()->getContents());
    foreach ($data->builds as $build) {
      foreach ($build->actions as $action) {
        if (isset($action->parameters)) {
          foreach ($action->parameters as $parameter) {
            if ($parameter->name === 'Drupal_JobID' && !empty($parameter->value) && strpos($parameter->value, $GLOBALS['base_url'] . ':') === 0) {
              if ($job = PiftCiJob::load((int) str_replace($GLOBALS['base_url'] . ':', '', $parameter->value))) {
                drush_log(dt('Found job !job_id', array('!job_id' => $job->job_id)));
                if ($job->ci_url === $url . '/' . $build->number || (strpos($job->ci_url, $url . '/') === 0 && str_replace($url . '/', '', $job->ci_url) > $build->number)) {
                  // Job is already updated or has a more recent build number.
                  break;
                }
                $job->status = PIFT_CI_STATUS_RUNNING;
                $job->message = t('Running');
                $job->ci_url = $url . '/' . $build->number;
                if ($job->save()) {
                  drush_log(dt('Saved job !job_id', array('!job_id' => $job->job_id)));
                }
                else {
                  drush_log(dt('Failed to save job !job_id', array('!job_id' => $job->job_id)));
                }
                break;
              }
              else {
                drush_log(dt('Failed to load CI job for @value', array('@value' => $parameter->value)));
              }
            }
          }
        }
      }
    }
  }
}

/**
 * Fetch DrupalCI results for each job.
 */
function drush_pift_cron_parse_results() {
  $query = new EntityFieldQuery();
  $query->entityCondition('entity_type', 'pift_ci_job')
    ->propertyCondition('status', PIFT_CI_STATUS_RUNNING);
  $result = $query->execute();
  if (isset($result['pift_ci_job'])) {
    foreach (entity_load('pift_ci_job', array_keys($result['pift_ci_job'])) as $queued_job) {
      $queued_job->requestResult();
    }
  }
}

/**
 * Queue daily & weekly jobs.
 */
function drush_pift_cron_daily() {
  $run_if_over = [
    // Run daily jobs if they have not run in the last 20 hours.
    PIFT_CI_TEST_BRANCH_DAILY => REQUEST_TIME - 60*60*20,
    // Run weekly jobs if not run in the last 6.5 days.
    PIFT_CI_TEST_BRANCH_WEEKLY => REQUEST_TIME - 60*60*24*6.5,
  ];
  $configurations_seen = [];
  $result = db_select('pift_ci_project')
    ->fields('pift_ci_project', array('nid', 'environment', 'release_nid', 'core_branch', 'testing'))
    ->condition('testing', array_keys($run_if_over))
    ->execute();
  foreach ($result as $row) {
    $release_wrapper = entity_metadata_wrapper('node', node_load($row->release_nid));
    $core_branch = pift_get_current_core($row->core_branch, $release_wrapper->field_release_category->value());

    // Catch duplicate configurations. For example, ~prerelease and ~supported
    // are sometimes the same thing.
    $configuration = $row->release_nid . '-' . $row->environment . '-' . $core_branch;
    if (isset($configurations_seen[$configuration])) {
      continue;
    }
    $configurations_seen[$configuration] = TRUE;

    // Check for existing job.
    $query = new EntityFieldQuery();
    $query->entityCondition('entity_type', 'pift_ci_job')
      ->propertyCondition('target_type', PIFT_CI_TYPE_BRANCH)
      ->propertyCondition('release_nid', $row->release_nid)
      ->propertyCondition('environment', $row->environment)
      ->propertyCondition('custom', FALSE)
      ->propertyOrderBy('updated', 'DESC')
      ->range(0, 1);
    if ($row->nid != PIFT_PID) {
      $query->propertyCondition('core_branch', $core_branch);
    }
    $jobs = $query->execute();
    if (isset($jobs['pift_ci_job'])) {
      $job = entity_load('pift_ci_job', array_keys($jobs['pift_ci_job']));
      $job = reset($job);

      if ($job->isRunning() || $job->created > $run_if_over[$row->testing]) {
        // Already queued, or a test has run in the last 20 hours, there is
        // nothing to do.
        continue;
      }

      // Queue a job.
      $data = new PiftCiJob(array(
        'reason' => 'Daily testing.',
        'target_type' => PIFT_CI_TYPE_BRANCH,
        'release_nid' => $row->release_nid,
        'environment' => $row->environment,
        'core_branch' => $row->core_branch,
        'build_iteration' => $job->build_iteration + 1,
      ));
    }
    else {
      // Make a fresh job.
      $data = new PiftCiJob(array(
        'reason' => 'Daily testing.',
        'target_type' => PIFT_CI_TYPE_BRANCH,
        'release_nid' => $row->release_nid,
        'environment' => $row->environment,
        'core_branch' => $row->core_branch,
      ));
    }
    $data->save();
  }
}

/**
 * Delete outdated results and data.
 */
function drush_pift_sweep_data() {
  // Remove coding standards results from closed issues.
  if ($job_ids = db_query('SELECT DISTINCT jc.job_id FROM {pift_ci_job_checkstyle} jc INNER JOIN {pift_ci_job} j ON j.job_id = jc.job_id AND j.target_type IN (:targets) INNER JOIN {field_data_field_issue_status} fis ON fis.entity_id = j.issue_nid AND fis.field_issue_status_value NOT IN (:open)', [':targets' => [PIFT_CI_TYPE_FILE, PIFT_CI_TYPE_MERGE_REQUEST], ':open' => project_issue_open_states()])->fetchCol()) {
    drush_log(dt('Cleaning issue code style results: @ids', ['@ids' => implode(', ', $job_ids)]));
    db_delete('pift_ci_job_checkstyle')->condition('job_id', $job_ids)->execute();
  }

  // … and from custom branch tests that have dropped off the project’s page.
  if ($job_ids = db_query('SELECT DISTINCT jc.job_id FROM {pift_ci_job_checkstyle} jc INNER JOIN {pift_ci_job} j ON j.job_id = jc.job_id AND j.target_type = :target AND j.custom = 1 AND created < :custom_time', [':target' => PIFT_CI_TYPE_BRANCH, ':custom_time' => REQUEST_TIME - PIFT_CUSTOM_TIME])->fetchCol()) {
    drush_log(dt('Cleaning custom branch code style results: @ids', ['@ids' => implode(', ', $job_ids)]));
    db_delete('pift_ci_job_checkstyle')->condition('job_id', $job_ids)->execute();
  }

  // … and from non-custom branch tests that are older than 24 hours and are
  // not the most recent results.
  $result = db_query('SELECT j.release_nid, j.environment, max(j.job_id) latest_job_id FROM {pift_ci_job_checkstyle} jc INNER JOIN {pift_ci_job} j ON j.job_id = jc.job_id AND j.target_type = :target AND j.custom = 0 GROUP BY j.release_nid, j.environment', [':target' => PIFT_CI_TYPE_BRANCH]);
  foreach ($result as $row) {
    if ($job_ids = db_query('SELECT DISTINCT jc.job_id FROM {pift_ci_job_checkstyle} jc INNER JOIN {pift_ci_job} j ON j.job_id = jc.job_id AND j.target_type = :target AND j.custom = 0 AND created < :time AND j.release_nid = :release_nid AND j.environment = :environment AND j.job_id < :latest_job_id', [':target' => PIFT_CI_TYPE_BRANCH, ':time' => REQUEST_TIME - 24*60*60, ':release_nid' => $row->release_nid, ':environment' => $row->environment, ':latest_job_id' => $row->latest_job_id])->fetchCol()) {
      drush_log(dt('Cleaning branch code style results for release @release_nid @environment: @ids', ['@release_nid' => $row->release_nid, '@environment' => $row->environment, '@ids' => implode(', ', $job_ids)]));
      db_delete('pift_ci_job_checkstyle')->condition('job_id', $job_ids)->execute();
    }
  }

  // Remove all jobs from dispatcher that are older than 180 days.
  if ($result = db_query("SELECT DISTINCT pcj.job_id FROM {pift_ci_job} pcj WHERE pcj.updated < UNIX_TIMESTAMP(DATE_SUB(NOW(), INTERVAL 180 DAY)) AND pcj.ci_url != '';")) {
    foreach ($result as $jobid) {
      $job = PiftCiJob::load($jobid->job_id);
      $ci_url = $job->ci_url;
      $job->purgeDispatcherJob();
      drush_log(dt('Deleted dispatcher results @ci_url from job #@job_id', ['@ci_url' => $ci_url, '@job_id' => $job->job_id]));
    }
  }

  // Remove all jobs from dispatcher that are passing branch tests that
  // have a newer passing test
  if ($result = db_query("SELECT DISTINCT pcj.job_id FROM {pift_ci_job} pcj
  LEFT JOIN (SELECT MAX(pcj.job_id) as 'job_id'
             FROM {pift_ci_job} pcj
             WHERE pcj.target_type = 'branch'
                   AND pcj.result = 'pass'
                   AND pcj.ci_url != ''
             GROUP BY pcj.release_nid, pcj.core_branch, pcj.environment ) AS pcj2 ON pcj2.job_id = pcj.job_id
WHERE pcj.target_type = 'branch'
      AND pcj.result = 'pass'
      AND pcj.ci_url != ''
      AND pcj2.job_id IS NULL
ORDER BY pcj.release_nid, pcj.job_id ASC;")) {
    foreach ($result as $jobid) {
      $job = PiftCiJob::load($jobid->job_id);
      $ci_url = $job->ci_url;
      $job->purgeDispatcherJob();
      drush_log(dt('Deleted dispatcher results @ci_url from job #@job_id', ['@ci_url' => $ci_url, '@job_id' => $job->job_id]));

    }
  }

  // Remove all test results from dispatcher for closed/fixed issues
  if ($result = db_query("SELECT DISTINCT pcj.job_id FROM {pift_ci_job} pcj
  LEFT JOIN {field_data_field_issue_status} fdfis ON fdfis.entity_id = pcj.issue_nid
WHERE pcj.target_type = 'file'
      AND fdfis.field_issue_status_value IN (3,5,6,7,17,18)
      AND pcj.ci_url != ''
      AND pcj.issue_nid IS NOT NULL
      AND pcj.updated < UNIX_TIMESTAMP(DATE_SUB(NOW(), INTERVAL 30 DAY));")) {
    foreach ($result as $jobid) {
      $job = PiftCiJob::load($jobid->job_id);
      $ci_url = $job->ci_url;
      $job->purgeDispatcherJob();
      drush_log(dt('Deleted dispatcher results @ci_url from job #@job_id', ['@ci_url' => $ci_url, '@job_id' => $job->job_id]));
    }
  }

  // Keep only the most recent RTBC patch job for a file/nid combo:
  if ($result = db_query("SELECT DISTINCT pcj.job_id FROM {pift_ci_job} pcj
  LEFT JOIN (SELECT MAX(pcj.job_id) as 'job_id'
             FROM {pift_ci_job} pcj
             WHERE pcj.target_type = 'file'
  AND pcj.reason = 'Automatic re-testing.'
  AND pcj.ci_url != ''
             GROUP BY pcj.issue_nid, pcj.file_id ) AS pcj2 ON pcj2.job_id = pcj.job_id
WHERE pcj.target_type = 'file'
  AND  pcj.reason = 'Automatic re-testing.'
  AND pcj.ci_url != ''
  AND pcj2.job_id IS NULL
ORDER BY pcj.release_nid, pcj.job_id ASC;")) {
    foreach ($result as $jobid) {
      $job = PiftCiJob::load($jobid->job_id);
      $ci_url = $job->ci_url;
      $job->purgeDispatcherJob();
      drush_log(dt('Deleted dispatcher results @ci_url from job #@job_id', ['@ci_url' => $ci_url, '@job_id' => $job->job_id]));
    }
  }

  // Remove issue test results from local site for closed issues over 180 days.
  $job_ids = db_query("SELECT DISTINCT pcj.job_id FROM {pift_ci_job} pcj INNER JOIN {pift_ci_job_result} pcjr ON pcjr.job_id = pcj.job_id INNER JOIN {field_data_field_issue_status} fdfis ON fdfis.entity_id = pcj.issue_nid AND fdfis.field_issue_status_value NOT IN (:open) WHERE pcj.target_type = :target AND pcj.issue_nid IS NOT NULL AND pcj.updated < UNIX_TIMESTAMP(DATE_SUB(NOW(), INTERVAL 180 DAY))", [
    ':open' => project_issue_open_states(),
    ':target' => PIFT_CI_TYPE_FILE,
  ])->fetchCol();
  foreach (array_chunk($job_ids, 50) as $job_id_chunk) {
    drush_log(dt('Deleting local results from jobs @job_ids', ['@job_ids' => implode(',', $job_id_chunk)]));
    db_delete('pift_ci_job_result')->condition('job_id', $job_id_chunk)->execute();
    db_delete('pift_ci_job_result_diff')->condition('job_id', $job_id_chunk)->execute();
    foreach (entity_load('pift_ci_job', $job_id_chunk) as $job) {
      $job->status = PIFT_CI_STATUS_SWEPT;
      $job->save(FALSE, TRUE);
    }
  }

  // Remove custom branch test results that have dropped off the project’s
  // page.
  $job_ids = db_query("SELECT DISTINCT pcj.job_id FROM pift_ci_job pcj WHERE pcj.target_type = :target AND pcj.updated < :custom_time AND pcj.status != :status_swept AND pcj.custom = 1", [
    ':target' => PIFT_CI_TYPE_BRANCH,
    ':custom_time' => REQUEST_TIME - PIFT_CUSTOM_TIME,
    ':status_swept' => PIFT_CI_STATUS_SWEPT,
  ])->fetchCol();
  foreach (array_chunk($job_ids, 50) as $job_id_chunk) {
    drush_log(dt('Deleting local results from jobs @job_ids', ['@job_ids' => implode(',', $job_id_chunk)]));
    db_delete('pift_ci_job_result')->condition('job_id', $job_id_chunk)->execute();
    foreach (entity_load('pift_ci_job', $job_id_chunk) as $job) {
      $job->status = PIFT_CI_STATUS_SWEPT;
      $job->save(FALSE, TRUE);
    }
  }

  // Remove non-custom branch test results that are older than 180 days and are
  // not the most recent results, and do not have a dependant branch job.
  $result = db_query('SELECT j.release_nid, j.environment, max(j.job_id) latest_job_id FROM {pift_ci_job} j WHERE j.target_type = :target AND j.custom = 0 GROUP BY j.release_nid, j.environment', [
    ':target' => PIFT_CI_TYPE_BRANCH,
  ]);
  foreach ($result as $row) {
    $job_ids = db_query('SELECT DISTINCT pcj.job_id FROM {pift_ci_job} pcj LEFT JOIN {pift_ci_job} pcj_diff ON pcj.job_id = pcj_diff.result_diff_job_id AND pcj_diff.status != :status_swept WHERE pcj.target_type = :target AND pcj.updated < UNIX_TIMESTAMP(DATE_SUB(NOW(), INTERVAL 180 DAY)) AND pcj.status != :status_swept AND pcj_diff.result_diff_job_id IS NULL AND pcj.custom = 0 AND pcj.release_nid = :release_nid AND pcj.environment = :environment AND pcj.job_id < :latest_job_id', [
      ':target' => PIFT_CI_TYPE_BRANCH,
      ':status_swept' => PIFT_CI_STATUS_SWEPT,
      ':release_nid' => $row->release_nid,
      ':environment' => $row->environment,
      ':latest_job_id' => $row->latest_job_id,
    ])->fetchCol();
    foreach (array_chunk($job_ids, 50) as $job_id_chunk) {
      drush_log(dt('Deleting local results from jobs @job_ids', ['@job_ids' => implode(',', $job_id_chunk)]));
      db_delete('pift_ci_job_result')->condition('job_id', $job_id_chunk)->execute();
      foreach (entity_load('pift_ci_job', $job_id_chunk) as $job) {
        $job->status = PIFT_CI_STATUS_SWEPT;
        $job->save(FALSE, TRUE);
      }
    }
  }
}

/**
 * Disable testing with a core label, and testing configurations using it.
 */
function drush_pift_disable_core_label($label) {
  $count = db_query('SELECT count(1) FROM {pift_ci_project} WHERE core_branch = :label AND testing != :disabled', [
    ':label' => $label,
    ':disabled' => PIFT_CI_TEST_DISABLED,
  ])->fetchField();
  if (!drush_confirm(dt('@count project testing configurations will be disabled', ['@count' => $count]))) {
    return;
  }
  db_update('pift_ci_label_map')
    ->condition('machine_label', $label)
    ->fields(['status' => 0])
    ->execute();
  db_update('pift_ci_project')
    ->condition('core_branch', $label)
    ->fields(['testing' => PIFT_CI_TEST_DISABLED])
    ->execute();
}

/**
 * Queue an individual branch test for a Git push.
 */
function drush_pift_queue_branch_test($project_name, $branch_name, $commit_id) {
  if (!($project = project_load($project_name))) {
    drush_set_error('PIFT_PROJECT_NOT_FOUND');
    return;
  }
  $repository = versioncontrol_project_repository_load($project->nid);
  if (!($release = project_release_get_release_for_vcs_label($project->nid, $branch_name))) {
    drush_set_error('PIFT_RELEASE_NOT_FOUND');
    return;
  }

  $repository = versioncontrol_project_repository_load($project->nid);
  $release_wrapper = entity_metadata_wrapper('node', $release);

  // We only support one drupalci.yml file in the root of the repository.
  $expected_path = $project->type === 'project_core' ? '/core/drupalci.yml' : '/drupalci.yml';
  $url = $repository->getUrlHandler()->getRepositoryViewUrl() . '/-/raw/' . $branch_name . $expected_path;
  try {
    $contents = (new \GuzzleHttp\Client())->get($url)->getBody();
    $query = db_select('pift_ci_config')
      ->condition('repo_id', $repository->repo_id)
      ->condition('branch', $branch_name)
      ->condition('file_path', $expected_path)
      ->condition('contents', $contents);
    $query->addExpression('1');
    if (!$query->execute()->fetchField()) {
      $result = db_merge('pift_ci_config')->key([
        'repo_id' => $repository->repo_id,
        'branch' => $branch_name,
        'file_path' => $expected_path,
      ])
      ->fields([
        'repo_name' => $repository->name,
        'file_type' => 'drupalci_testing_config',
        'updated' => REQUEST_TIME,
        'contents' => $contents,
      ])->execute();
    }
  }
  catch (Exception $e) {
    if (!($e instanceof \GuzzleHttp\Exception\ClientException && $e->getCode() == 404)) {
      watchdog('pift_ci', 'Failure to update drupalci.yml: @error', ['@error' => $e->getMessage()], WATCHDOG_WARNING);
    }
  }

  // Look for testable dev release branches.
  foreach (pift_ci_project_info(project_release_get_release_project_nid($release)) as $environment => $versions) {
    if (isset($versions[$release->nid])) {
      $configurations_seen = [];
      foreach ($versions[$release->nid] as $core_branch => $testing) {
        // Catch duplicate configurations. For example, ~prerelease and
        // ~supported are sometimes the same thing.
        $configuration = $environment . '-' . pift_get_current_core($core_branch, $release_wrapper->field_release_category->value());
        if (!isset($configurations_seen[$configuration]) && ($testing === PIFT_CI_TEST_BRANCH_COMMIT || $testing === PIFT_CI_TEST_ISSUE)) {
          drush_log(dt('Queueing test for !configuration', ['!configuration' => $configuration]));
          $configurations_seen[$configuration] = TRUE;
          $job = new PiftCiJob([
            'reason' => '@commit_id pushed to Git.',
            'target_type' => PIFT_CI_TYPE_BRANCH,
            'release_nid' => $release->nid,
            'commit_id' => $commit_id,
            'environment' => $environment,
            'core_branch' => $core_branch,
          ]);
          $job->save();
        }
      }
    }
  }
}
